@@ -16,30 +16,30 @@ extern options opts;
 buffer ibuf;			/* Input buffer. */
 enum {				/* Server data responses to be parsed;
 				 * regular expressions index. */
-	DATA_RESPONSE_TAGGED,
-	DATA_RESPONSE_CAPABILITY,
-	DATA_RESPONSE_AUTHENTICATE,
-	DATA_RESPONSE_NAMESPACE,
-	DATA_RESPONSE_STATUS,
-	DATA_RESPONSE_STATUS_MESSAGES,
-	DATA_RESPONSE_STATUS_RECENT,
-	DATA_RESPONSE_STATUS_UNSEEN,
-	DATA_RESPONSE_STATUS_UIDNEXT,
-	DATA_RESPONSE_EXAMINE_EXISTS,
-	DATA_RESPONSE_EXAMINE_RECENT,
-	DATA_RESPONSE_LIST,
-	DATA_RESPONSE_SEARCH,
-	DATA_RESPONSE_FETCH,
-	DATA_RESPONSE_FETCH_FLAGS,
-	DATA_RESPONSE_FETCH_DATE,
-	DATA_RESPONSE_FETCH_SIZE,
-	DATA_RESPONSE_FETCH_STRUCTURE,
-	DATA_RESPONSE_FETCH_BODY,
-	DATA_RESPONSE_IDLE,
+	RESPONSE_TAGGED,
+	RESPONSE_CAPABILITY,
+	RESPONSE_AUTHENTICATE,
+	RESPONSE_NAMESPACE,
+	RESPONSE_STATUS,
+	RESPONSE_STATUS_MESSAGES,
+	RESPONSE_STATUS_RECENT,
+	RESPONSE_STATUS_UNSEEN,
+	RESPONSE_STATUS_UIDNEXT,
+	RESPONSE_EXAMINE_EXISTS,
+	RESPONSE_EXAMINE_RECENT,
+	RESPONSE_LIST,
+	RESPONSE_SEARCH,
+	RESPONSE_FETCH,
+	RESPONSE_FETCH_FLAGS,
+	RESPONSE_FETCH_DATE,
+	RESPONSE_FETCH_SIZE,
+	RESPONSE_FETCH_STRUCTURE,
+	RESPONSE_FETCH_BODY,
+	RESPONSE_IDLE,
 };
 regexp responses[] = {		/* Server data responses to be parsed;
 				 * regular expressions patterns. */
-	{ "([[:xdigit:]]{4,4}) (OK|NO|BAD) [[:print:]]*\r\n", NULL, 0, NULL },
+	{ "([[:xdigit:]]{4,4}) (OK|NO|BAD) [^[:cntrl:]]*\r\n", NULL, 0, NULL },
 	{ "\\* CAPABILITY ([[:print:]]*)\r\n", NULL, 0, NULL },
 	{ "\\+ ([[:graph:]]*)\r\n", NULL, 0, NULL },
 	{ "\\* NAMESPACE (NIL|\\(\\(\"([[:graph:]]*)\" \"([[:print:]])\"\\)"
@@ -89,7 +89,7 @@ receive_response(session *ssn, char *buf, long timeout, int timeoutfail)
 		return -1;
 
 
-	if (opts.debug > 0) {
+	if (opts.debug) {
 		int i;
 		
 		debug("getting response (%d):\n\n", ssn->socket);
@@ -114,31 +114,31 @@ check_tag(char *buf, session *ssn, int tag)
 	char t[4 + 1];
 	regexp *re;
 
-	r = STATUS_RESPONSE_NONE;
+	r = STATUS_NONE;
 
 	snprintf(t, sizeof(t), "%04X", tag);
 
-	re = &responses[DATA_RESPONSE_TAGGED];
+	re = &responses[RESPONSE_TAGGED];
 
 	if (!regexec(re->preg, buf, re->nmatch, re->pmatch, 0)) {
 		if (!strncasecmp(buf + re->pmatch[1].rm_so, t,
 		    strlen(t))) {
 			if (!strncasecmp(buf + re->pmatch[2].rm_so,
 			    "OK", strlen("OK")))
-				r = STATUS_RESPONSE_OK;
+				r = STATUS_OK;
 			else if (!strncasecmp(buf + re->pmatch[2].rm_so,
 			    "NO", strlen("NO")))
-				r = STATUS_RESPONSE_NO;
+				r = STATUS_NO;
 			else if (!strncasecmp(buf + re->pmatch[2].rm_so,
 			    "BAD", strlen("BAD")))
-				r = STATUS_RESPONSE_BAD;
+				r = STATUS_BAD;
 		}
 	}
 
-	if (r != STATUS_RESPONSE_NONE)
+	if (r != STATUS_NONE)
 		verbose("S (%d): %s", ssn->socket, buf + re->pmatch[0].rm_so);
 
-	if (r == STATUS_RESPONSE_NO || r == STATUS_RESPONSE_BAD)
+	if (r == STATUS_NO || r == STATUS_BAD)
 		error("IMAP (%d): %s", ssn->socket, buf + re->pmatch[0].rm_so);
 
 	return r;
@@ -219,17 +219,18 @@ response_generic(session *ssn, int tag)
 
 	do {
 		buffer_check(&ibuf, ibuf.len + INPUT_BUF);
-		if ((n = receive_response(ssn, ibuf.data + ibuf.len, 0, 1)) == -1)
+		if ((n = receive_response(ssn, ibuf.data + ibuf.len, 0, 1)) ==
+		    -1)
 			return -1;
 		ibuf.len += n;
 
 		if (check_bye(ibuf.data))
-			return -1;
-	} while ((r = check_tag(ibuf.data, ssn, tag)) == STATUS_RESPONSE_NONE);
+			return STATUS_BYE;
+	} while ((r = check_tag(ibuf.data, ssn, tag)) == STATUS_NONE);
 
-	if (r == STATUS_RESPONSE_NO &&
+	if (r == STATUS_NO &&
 	    (check_trycreate(ibuf.data) || get_option_boolean("create")))
-		return STATUS_RESPONSE_TRYCREATE;
+		return STATUS_TRYCREATE;
 
 	return r;
 }
@@ -239,23 +240,33 @@ response_generic(session *ssn, int tag)
  * Get server data and make sure there is a continuation response inside them.
  */
 int
-response_continuation(session *ssn)
+response_continuation(session *ssn, int tag)
 {
+	int r;
 	ssize_t n;
 
 	buffer_reset(&ibuf);
 
 	do {
 		buffer_check(&ibuf, ibuf.len + INPUT_BUF);
-		if ((n = receive_response(ssn, ibuf.data + ibuf.len, 0, 1)) == -1)
+		if ((n = receive_response(ssn, ibuf.data + ibuf.len, 0, 1)) ==
+		    -1)
 			return -1;
 		ibuf.len += n;
 
 		if (check_bye(ibuf.data))
-			return -1;
-	} while (!check_continuation(ibuf.data));
+			return STATUS_BYE;
+	} while ((r = check_tag(ibuf.data, ssn, tag)) == STATUS_NONE &&
+	    !check_continuation(ibuf.data));
 
-	return STATUS_RESPONSE_CONTINUE;
+	if (r == STATUS_NO &&
+	    (check_trycreate(ibuf.data) || get_option_boolean("create")))
+		return STATUS_TRYCREATE;
+
+	if (r == STATUS_NONE)
+		return STATUS_CONTINUE;
+
+	return r;
 }
 
 
@@ -274,12 +285,12 @@ response_greeting(session *ssn)
 	verbose("S (%d): %s", ssn->socket, ibuf.data);
 
 	if (check_bye(ibuf.data))
-		return -1;
+		return STATUS_BYE;
 
 	if (check_preauth(ibuf.data))
-		return STATUS_RESPONSE_PREAUTH;
+		return STATUS_PREAUTH;
 
-	return STATUS_RESPONSE_NONE;
+	return STATUS_NONE;
 }
 
 
@@ -293,12 +304,13 @@ response_capability(session *ssn, int tag)
 	char *s;
 	regexp *re;
 
-	if ((r = response_generic(ssn, tag)) == -1)
-		return -1;
+	r = response_generic(ssn, tag);
+	if (r == -1 || r == STATUS_BYE)
+		return r;
 
 	ssn->protocol = PROTOCOL_NONE;
 
-	re = &responses[DATA_RESPONSE_CAPABILITY];
+	re = &responses[RESPONSE_CAPABILITY];
 
 	if (!regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0)) {
 		s = xstrndup(ibuf.data + re->pmatch[1].rm_so,
@@ -318,14 +330,10 @@ response_capability(session *ssn, int tag)
 
 		if (xstrcasestr(s, "NAMESPACE"))
 			ssn->capabilities |= CAPABILITY_NAMESPACE;
-#ifndef NO_CRAMMD5
 		if (xstrcasestr(s, "AUTH=CRAM-MD5"))
 			ssn->capabilities |= CAPABILITY_CRAMMD5;
-#endif
-#ifndef NO_SSLTLS
 		if (xstrcasestr(s, "STARTTLS"))
 			ssn->capabilities |= CAPABILITY_STARTTLS;
-#endif
 		if (xstrcasestr(s, "CHILDREN"))
 			ssn->capabilities |= CAPABILITY_CHILDREN;
 
@@ -339,7 +347,6 @@ response_capability(session *ssn, int tag)
 }
 
 
-#ifndef NO_CRAMMD5
 /*
  * Process the data that server sent due to IMAP AUTHENTICATE client request.
  */
@@ -349,16 +356,16 @@ response_authenticate(session *ssn, int tag, unsigned char **cont)
 	int r;
 	regexp *re;
 
-	re = &responses[DATA_RESPONSE_AUTHENTICATE];
+	re = &responses[RESPONSE_AUTHENTICATE];
 
-	if ((r = response_continuation(ssn)) == STATUS_RESPONSE_CONTINUE &&
+	if ((r = response_continuation(ssn, tag)) == STATUS_CONTINUE &&
 	    !regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0))
-		*cont = (unsigned char *)xstrndup(ibuf.data + re->pmatch[1].rm_so,
-		    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
+		*cont = (unsigned char *)xstrndup(ibuf.data +
+		    re->pmatch[1].rm_so, re->pmatch[1].rm_eo -
+		    re->pmatch[1].rm_so);
 
 	return r;
 }
-#endif
 
 
 /*
@@ -370,13 +377,14 @@ response_namespace(session *ssn, int tag)
 	int r, n;
 	regexp *re;
 
-	if ((r = response_generic(ssn, tag)) == -1)
-		return -1;
+	r = response_generic(ssn, tag);
+	if (r == -1 || r == STATUS_BYE)
+		return r;
 
 	ssn->ns.prefix = NULL;
 	ssn->ns.delim = '\0';
 
-	re = &responses[DATA_RESPONSE_NAMESPACE];
+	re = &responses[RESPONSE_NAMESPACE];
 
 	if (!regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0)) {
 		n = re->pmatch[2].rm_eo - re->pmatch[2].rm_so;
@@ -403,28 +411,29 @@ response_status(session *ssn, int tag, unsigned int *exist,
 	char *s;
 	regexp *re;
 
-	if ((r = response_generic(ssn, tag)) == -1)
-		return -1;
+	r = response_generic(ssn, tag);
+	if (r == -1 || r == STATUS_BYE)
+		return r;
 
-	re = &responses[DATA_RESPONSE_STATUS];
+	re = &responses[RESPONSE_STATUS];
 
 	if (!regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0)) {
 		s = xstrndup(ibuf.data + re->pmatch[1].rm_so,
 		    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
 
-		re = &responses[DATA_RESPONSE_STATUS_MESSAGES];
+		re = &responses[RESPONSE_STATUS_MESSAGES];
 		if (!regexec(re->preg, s, re->nmatch, re->pmatch, 0))
 			*exist = strtol(s + re->pmatch[1].rm_so, NULL, 10);
 
-		re = &responses[DATA_RESPONSE_STATUS_RECENT];
+		re = &responses[RESPONSE_STATUS_RECENT];
 		if (!regexec(re->preg, s, re->nmatch, re->pmatch, 0))
 			*recent = strtol(s + re->pmatch[1].rm_so, NULL, 10);
 
-		re = &responses[DATA_RESPONSE_STATUS_UNSEEN];
+		re = &responses[RESPONSE_STATUS_UNSEEN];
 		if (!regexec(re->preg, s, re->nmatch, re->pmatch, 0))
 			*unseen = strtol(s + re->pmatch[1].rm_so, NULL, 10);
 
-		re = &responses[DATA_RESPONSE_STATUS_UIDNEXT];
+		re = &responses[RESPONSE_STATUS_UIDNEXT];
 		if (!regexec(re->preg, s, re->nmatch, re->pmatch, 0))
 			*uidnext = strtol(s + re->pmatch[1].rm_so, NULL, 10);
 
@@ -445,14 +454,15 @@ response_examine(session *ssn, int tag, unsigned int *exist,
 	int r;
 	regexp *re;
 
-	if ((r = response_generic(ssn, tag)) == -1)
-		return -1;
+	r = response_generic(ssn, tag);
+	if (r == -1 || r == STATUS_BYE)
+		return r;
 
-	re = &responses[DATA_RESPONSE_EXAMINE_EXISTS];
+	re = &responses[RESPONSE_EXAMINE_EXISTS];
 	if (!regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0))
 		*exist = strtol(ibuf.data + re->pmatch[1].rm_so, NULL, 10);
 
-	re = &responses[DATA_RESPONSE_EXAMINE_RECENT];
+	re = &responses[RESPONSE_EXAMINE_RECENT];
 	if (!regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0))
 		*recent = strtol(ibuf.data + re->pmatch[1].rm_so, NULL, 10);
 
@@ -468,11 +478,12 @@ response_select(session *ssn, int tag)
 {
 	int r;
 
-	if ((r = response_generic(ssn, tag)) == -1)
-		return -1;
+	r = response_generic(ssn, tag);
+	if (r == -1 || r == STATUS_BYE)
+		return r;
 
 	if (xstrcasestr(ibuf.data, "[READ-ONLY]"))
-		return STATUS_RESPONSE_READONLY;
+		return STATUS_READONLY;
 
 	return r;
 }
@@ -490,14 +501,15 @@ response_list(session *ssn, int tag, char **mboxs, char **folders)
 	const char *v;
 	regexp *re;
 
-	if ((r = response_generic(ssn, tag)) == -1)
-		return -1;
+	r = response_generic(ssn, tag);
+	if (r == -1 || r == STATUS_BYE)
+		return r;
 
 	m = *mboxs = (char *)xmalloc((ibuf.len + 1) * sizeof(char));
 	f = *folders = (char *)xmalloc((ibuf.len + 1) * sizeof(char));
 	*m = *f = '\0';
 
-	re = &responses[DATA_RESPONSE_LIST];
+	re = &responses[RESPONSE_LIST];
 
 	b = ibuf.data;
 	while (!regexec(re->preg, b, re->nmatch, re->pmatch, 0)) {
@@ -557,10 +569,11 @@ response_search(session *ssn, int tag, char **mesgs)
 	regexp *re;
 	char *b, *m;
 
-	if ((r = response_generic(ssn, tag)) == -1)
-		return -1;
+	r = response_generic(ssn, tag);
+	if (r == -1 || r == STATUS_BYE)
+		return r;
 
-	re = &responses[DATA_RESPONSE_SEARCH];
+	re = &responses[RESPONSE_SEARCH];
 
 	b = ibuf.data;
 	m = NULL;
@@ -571,7 +584,8 @@ response_search(session *ssn, int tag, char **mesgs)
 			*m = '\0';
 		}
 
-		min = (unsigned int)(re->pmatch[1].rm_eo - re->pmatch[1].rm_so) < ibuf.len ?
+		min = (unsigned int)(re->pmatch[1].rm_eo -
+		    re->pmatch[1].rm_so) < ibuf.len ?
 		    (unsigned int)(re->pmatch[1].rm_eo - re->pmatch[1].rm_so) :
 		    ibuf.len;
 
@@ -597,25 +611,26 @@ response_fetchfast(session *ssn, int tag, char **flags, char **date,
 	char *s;
 	regexp *re;
 
-	if ((r = response_generic(ssn, tag)) == -1)
-		return -1;
+	r = response_generic(ssn, tag);
+	if (r == -1 || r == STATUS_BYE)
+		return r;
 
-	re = &responses[DATA_RESPONSE_FETCH];
+	re = &responses[RESPONSE_FETCH];
 	if (!regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0)) { 
 		s = xstrndup(ibuf.data + re->pmatch[1].rm_so,
 		    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
 
-		re = &responses[DATA_RESPONSE_FETCH_FLAGS];
+		re = &responses[RESPONSE_FETCH_FLAGS];
 		if (!regexec(re->preg, s, re->nmatch, re->pmatch, 0))
 			*flags = xstrndup(s + re->pmatch[1].rm_so,
 			    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
 
-		re = &responses[DATA_RESPONSE_FETCH_DATE];
+		re = &responses[RESPONSE_FETCH_DATE];
 		if (!regexec(re->preg, s, re->nmatch, re->pmatch, 0))
 			*date = xstrndup(s + re->pmatch[1].rm_so,
 			    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
 
-		re = &responses[DATA_RESPONSE_FETCH_SIZE];
+		re = &responses[RESPONSE_FETCH_SIZE];
 		if (!regexec(re->preg, s, re->nmatch, re->pmatch, 0))
 			*size = xstrndup(s + re->pmatch[1].rm_so,
 			    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
@@ -637,15 +652,16 @@ response_fetchflags(session *ssn, int tag, char **flags)
 	char *s;
 	regexp *re;
 
-	if ((r = response_generic(ssn, tag)) == -1)
-		return -1;
+	r = response_generic(ssn, tag);
+	if (r == -1 || r == STATUS_BYE)
+		return r;
 
-	re = &responses[DATA_RESPONSE_FETCH];
+	re = &responses[RESPONSE_FETCH];
 	if (!regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0)) { 
 		s = xstrndup(ibuf.data + re->pmatch[1].rm_so,
 		    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
 
-		re = &responses[DATA_RESPONSE_FETCH_FLAGS];
+		re = &responses[RESPONSE_FETCH_FLAGS];
 		if (!regexec(re->preg, s, re->nmatch, re->pmatch, 0))
 			*flags = xstrndup(s + re->pmatch[1].rm_so,
 			    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
@@ -668,15 +684,16 @@ response_fetchdate(session *ssn, int tag, char **date)
 	char *s;
 	regexp *re;
 
-	if ((r = response_generic(ssn, tag)) == -1)
-		return -1;
+	r = response_generic(ssn, tag);
+	if (r == -1 || r == STATUS_BYE)
+		return r;
 
-	re = &responses[DATA_RESPONSE_FETCH];
+	re = &responses[RESPONSE_FETCH];
 	if (!regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0)) { 
 		s = xstrndup(ibuf.data + re->pmatch[1].rm_so,
 		    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
 
-		re = &responses[DATA_RESPONSE_FETCH_DATE];
+		re = &responses[RESPONSE_FETCH_DATE];
 		if (!regexec(re->preg, s, re->nmatch, re->pmatch, 0))
 			*date = xstrndup(s + re->pmatch[1].rm_so,
 			    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
@@ -699,15 +716,16 @@ response_fetchsize(session *ssn, int tag, char **size)
 	char *s;
 	regexp *re;
 
-	if ((r = response_generic(ssn, tag)) == -1)
-		return -1;
+	r = response_generic(ssn, tag);
+	if (r == -1 || r == STATUS_BYE)
+		return r;
 
-	re = &responses[DATA_RESPONSE_FETCH];
+	re = &responses[RESPONSE_FETCH];
 	if (!regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0)) { 
 		s = xstrndup(ibuf.data + re->pmatch[1].rm_so,
 		    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
 
-		re = &responses[DATA_RESPONSE_FETCH_SIZE];
+		re = &responses[RESPONSE_FETCH_SIZE];
 		if (!regexec(re->preg, s, re->nmatch, re->pmatch, 0))
 			*size = xstrndup(s + re->pmatch[1].rm_so,
 			    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
@@ -730,15 +748,16 @@ response_fetchstructure(session *ssn, int tag, char **structure)
 	char *s;
 	regexp *re;
 
-	if ((r = response_generic(ssn, tag)) == -1)
-		return -1;
+	r = response_generic(ssn, tag);
+	if (r == -1 || r == STATUS_BYE)
+		return r;
 
-	re = &responses[DATA_RESPONSE_FETCH];
+	re = &responses[RESPONSE_FETCH];
 	if (!regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0)) { 
 		s = xstrndup(ibuf.data + re->pmatch[1].rm_so,
 		    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
 
-		re = &responses[DATA_RESPONSE_FETCH_STRUCTURE];
+		re = &responses[RESPONSE_FETCH_STRUCTURE];
 		if (!regexec(re->preg, s, re->nmatch, re->pmatch, 0)) {
 			*structure = xstrndup(s + re->pmatch[1].rm_so,
 			    re->pmatch[1].rm_eo - re->pmatch[1].rm_so);
@@ -771,11 +790,12 @@ response_fetchbody(session *ssn, int tag, char **body, size_t *len)
 	match = -1;
 	offset = 0;
 	
-	re = &responses[DATA_RESPONSE_FETCH_BODY];
+	re = &responses[RESPONSE_FETCH_BODY];
 
 	do {
 		buffer_check(&ibuf, ibuf.len + INPUT_BUF);
-		if ((n = receive_response(ssn, ibuf.data + ibuf.len, 0, 1)) == -1)
+		if ((n = receive_response(ssn, ibuf.data + ibuf.len, 0, 1)) ==
+		    -1)
 			return -1;
 		ibuf.len += n;
 
@@ -793,10 +813,10 @@ response_fetchbody(session *ssn, int tag, char **body, size_t *len)
 
 		if (offset != 0 && ibuf.len >= offset) {
 			if (check_bye(ibuf.data + offset))
-				return -1;
+				return STATUS_BYE;
 		}
 	} while (ibuf.len < offset || (r = check_tag(ibuf.data + offset, ssn,
-	    tag)) == STATUS_RESPONSE_NONE);
+	    tag)) == STATUS_NONE);
 
 	if (match == 0) {
 		if (re->pmatch[2].rm_so != -1 &&
@@ -820,7 +840,10 @@ response_idle(session *ssn, int tag)
 {
 	regexp *re;
 
-	re = &responses[DATA_RESPONSE_IDLE];
+	if (tag == -1)
+		return -1;
+
+	re = &responses[RESPONSE_IDLE];
 
 	do {
 		buffer_reset(&ibuf);
@@ -831,16 +854,16 @@ response_idle(session *ssn, int tag)
 			return -1;
 			break; /* NOTREACHED */
 		case 0:
-			return STATUS_RESPONSE_TIMEOUT;
+			return STATUS_TIMEOUT;
 			break; /* NOTREACHED */
 		}
 
 		verbose("S (%d): %s", ssn->socket, ibuf.data);
 
 		if (check_bye(ibuf.data))
-			return -1;
+			return STATUS_BYE;
 
 	} while (regexec(re->preg, ibuf.data, re->nmatch, re->pmatch, 0));
 
-	return STATUS_RESPONSE_UNTAGGED;
+	return STATUS_UNTAGGED;
 }
