@@ -10,38 +10,28 @@
 #include <sys/time.h>
 #include <sys/select.h>
 
-#include "imapfilter.h"
-#include "session.h"
-
-#ifndef NO_SSLTLS
 #include <openssl/ssl.h>
 #include <openssl/err.h>
-#endif
+
+#include "imapfilter.h"
+#include "session.h"
 
 
 /*
  * Connect to mail server.
  */
 int
-open_connection(session *ssn, const char *server, const char *port,
-    const char *protocol)
+open_connection(session *ssn)
 {
 	struct addrinfo hints, *res, *ressave;
 	int n, sockfd;
 
-#ifdef NO_SSLTLS
-	if (protocol) {
-		error("SSL not supported by this build\n");
-		return -1;
-	}
-#endif
-
 	memset(&hints, 0, sizeof(struct addrinfo));
 
 	hints.ai_family = AF_UNSPEC;
 	hints.ai_socktype = SOCK_STREAM;
 
-	n = getaddrinfo(server, port, &hints, &res);
+	n = getaddrinfo(ssn->server, ssn->port, &hints, &res);
 
 	if (n < 0) {
 		error("gettaddrinfo; %s\n", gai_strerror(n));
@@ -70,62 +60,60 @@ open_connection(session *ssn, const char *server, const char *port,
 
 	if (sockfd == -1) {
 		error("error while initiating connection to %s at port %s\n",
-		    server, port);
+		    ssn->server, ssn->port);
 		return -1;
 	}
 
 	ssn->socket = sockfd;
 
-#ifndef NO_SSLTLS
-	if (protocol) {
-		if (open_secure_connection(ssn, server, port, protocol) == -1) {
+	if (ssn->sslproto) {
+		if (open_secure_connection(ssn) == -1) {
 			close_connection(ssn);
 			return -1;
 		}
 	}
-#endif
 
 	return ssn->socket;
 }
 
 
-#ifndef NO_SSLTLS
 /*
  * Initialize SSL/TLS connection.
  */
 int
-open_secure_connection(session *ssn, const char *server, const char *port,
-    const char *protocol)
+open_secure_connection(session *ssn)
 {
 	int r, e;
 	SSL_CTX *ctx;
-	SSL_METHOD *method;
+	const SSL_METHOD *method;
 
 	method = NULL;
 
-	if (!strncasecmp(protocol, "tls1", 4))
-		method = TLSv1_client_method();
-	else if (!strncasecmp(protocol, "ssl3", 4) ||
-	    !strncasecmp(protocol, "ssl2", 4))
+	if (ssn->sslproto && (!strncasecmp(ssn->sslproto, "ssl3", 4) ||
+	    !strncasecmp(ssn->sslproto, "ssl2", 4)))
 		method = SSLv23_client_method();
+	else
+		method = TLSv1_client_method();
 
 	if (!(ctx = SSL_CTX_new(method)))
 		goto fail;
 
-	if (!(ssn->ssl = SSL_new(ctx)))
+	if (!(ssn->sslconn = SSL_new(ctx)))
 		goto fail;
 
-	SSL_set_fd(ssn->ssl, ssn->socket);
+	SSL_set_fd(ssn->sslconn, ssn->socket);
 
 	for (;;) {
-		if ((r = SSL_connect(ssn->ssl)) > 0)
+		if ((r = SSL_connect(ssn->sslconn)) > 0)
 			break;
 
-		switch (SSL_get_error(ssn->ssl, r)) {
+		switch (SSL_get_error(ssn->sslconn, r)) {
 		case SSL_ERROR_ZERO_RETURN:
 			error("initiating SSL connection to %s; the "
-			    "connection has been closed cleanly\n", server);
+			    "connection has been closed cleanly\n",
+			    ssn->server);
 			goto fail;
+		case SSL_ERROR_NONE:
 		case SSL_ERROR_WANT_CONNECT:
 		case SSL_ERROR_WANT_ACCEPT:
 		case SSL_ERROR_WANT_X509_LOOKUP:
@@ -134,19 +122,23 @@ open_secure_connection(session *ssn, const char *server, const char *port,
 			break;
 		case SSL_ERROR_SYSCALL:
 			e = ERR_get_error();
-			if (e == 0)
-				error("initiating SSL connection to %s; EOF "
-				    "in violation of the protocol\n", server);
-			else if (e == -1)
+			if (e == 0 && r == 0)
+				error("initiating SSL connection to %s; EOF in "
+				    "violation of the protocol\n", ssn->server);
+			else if (e == 0 && r == -1)
 				error("initiating SSL connection to %s; %s\n",
-				    server, strerror(errno));
+				    ssn->server, strerror(errno));
+			else
+				error("initiating SSL connection to %s; %s\n",
+				    ssn->server, ERR_error_string(e, NULL));
 			goto fail;
 		case SSL_ERROR_SSL:
-			error("initiating SSL connection to %s; %s\n", server,
-			    ERR_error_string(ERR_get_error(), NULL));
+			error("initiating SSL connection to %s; %s\n",
+			    ssn->server, ERR_error_string(ERR_get_error(),
+			    NULL));
 			goto fail;
 		default:
-			goto fail;
+			break;
 		}
 	}
 	if (get_option_boolean("certificates") && get_cert(ssn) == -1)
@@ -157,12 +149,11 @@ open_secure_connection(session *ssn, const char *server, const char *port,
 	return 0;
 
 fail:
-	ssn->ssl = NULL;
+	ssn->sslconn = NULL;
 	SSL_CTX_free(ctx);
 
 	return -1;
 }
-#endif				/* NO_SSLTLS */
 
 
 /*
@@ -175,9 +166,7 @@ close_connection(session *ssn)
 
 	r = 0;
 
-#ifndef NO_SSLTLS
 	close_secure_connection(ssn);
-#endif
 
 	if (ssn->socket != -1) {
 		r = close(ssn->socket);
@@ -190,7 +179,6 @@ close_connection(session *ssn)
 }
 
 
-#ifndef NO_SSLTLS
 /*
  * Shutdown SSL/TLS connection.
  */
@@ -198,15 +186,14 @@ int
 close_secure_connection(session *ssn)
 {
 
-	if (ssn->ssl) {
-		SSL_shutdown(ssn->ssl);
-		SSL_free(ssn->ssl);
-		ssn->ssl = NULL;
+	if (ssn->sslconn) {
+		SSL_shutdown(ssn->sslconn);
+		SSL_free(ssn->sslconn);
+		ssn->sslconn = NULL;
 	}
 
 	return 0;
 }
-#endif
 
 
 /*
@@ -238,9 +225,8 @@ socket_read(session *ssn, char *buf, size_t len, long timeout, int timeoutfail)
 	FD_ZERO(&fds);
 	FD_SET(ssn->socket, &fds);
  
-#ifndef NO_SSLTLS
-	if (ssn->ssl) {
-		if (SSL_pending(ssn->ssl) > 0 ||
+	if (ssn->sslconn) {
+		if (SSL_pending(ssn->sslconn) > 0 ||
 		    ((s = select(ssn->socket + 1, &fds, NULL, NULL, tvp)) > 0 &&
 		    FD_ISSET(ssn->socket, &fds))) {
 			r = socket_secure_read(ssn, buf, len);
@@ -248,9 +234,7 @@ socket_read(session *ssn, char *buf, size_t len, long timeout, int timeoutfail)
 			if (r <= 0)
 				goto fail;
 		}
-	} else
-#endif
-	{
+	} else {
 		if ((s = select(ssn->socket + 1, &fds, NULL, NULL, tvp)) > 0 &&
 		    FD_ISSET(ssn->socket, &fds)) {
 			r = read(ssn->socket, buf, len);
@@ -281,7 +265,6 @@ fail:
 }
 
 
-#ifndef NO_SSLTLS
 /*
  * Read data from a TLS/SSL connection.
  */
@@ -291,16 +274,15 @@ socket_secure_read(session *ssn, char *buf, size_t len)
 	int r, e;
 
 	for (;;) {
-		r = (ssize_t) SSL_read(ssn->ssl, buf, len);
-
-		if (r > 0)
+		if ((r = (ssize_t) SSL_read(ssn->sslconn, buf, len)) > 0)
 			break;
 
-		switch (SSL_get_error(ssn->ssl, r)) {
+		switch (SSL_get_error(ssn->sslconn, r)) {
 		case SSL_ERROR_ZERO_RETURN:
-			error("reading data; the connection has been closed "
-			    "cleanly\n");
-			return -1;
+			error("reading data through SSL; the connection has "
+			    "been closed cleanly\n");
+			goto fail;
+		case SSL_ERROR_NONE:
 		case SSL_ERROR_WANT_READ:
 		case SSL_ERROR_WANT_WRITE:
 		case SSL_ERROR_WANT_CONNECT:
@@ -309,24 +291,33 @@ socket_secure_read(session *ssn, char *buf, size_t len)
 			break;
 		case SSL_ERROR_SYSCALL:
 			e = ERR_get_error();
-			if (e == 0)
-				error("reading data; EOF in violation of the "
-				    "protocol\n");
-			else if (e == -1)
-				error("reading data; %s\n", strerror(errno));
-			return -1;
+			if (e == 0 && r == 0)
+				error("reading data through SSL; EOF in "
+				    "violation of the protocol\n");
+			else if (e == 0 && r == -1)
+				error("reading data through SSL; %s\n",
+				    strerror(errno));
+			else
+				error("reading data through SSL; %s\n",
+				    ERR_error_string(e, NULL));
+			goto fail;
 		case SSL_ERROR_SSL:
-			error("reading data; %s\n",
+			error("reading data through SSL; %s\n",
 			    ERR_error_string(ERR_get_error(), NULL));
-			return -1;
+			goto fail;
 		default:
-			return -1;
+			break;
 		}
 	}
 
 	return r;
+fail:
+	SSL_set_shutdown(ssn->sslconn, SSL_SENT_SHUTDOWN |
+	    SSL_RECEIVED_SHUTDOWN);
+
+	return -1;
+
 }
-#endif
 
 
 /*
@@ -336,10 +327,10 @@ ssize_t
 socket_write(session *ssn, const char *buf, size_t len)
 {
 	int s;
-	ssize_t w, wt;
+	ssize_t r, t;
 	fd_set fds;
 
-	w = wt = 0;
+	r = t = 0;
 	s = 1;
 
 	FD_ZERO(&fds);
@@ -348,30 +339,27 @@ socket_write(session *ssn, const char *buf, size_t len)
 	while (len) {
 		if ((s = select(ssn->socket + 1, NULL, &fds, NULL, NULL) > 0 &&
 		    FD_ISSET(ssn->socket, &fds))) {
-#ifndef NO_SSLTLS
-			if (ssn->ssl) {
-				w = socket_secure_write(ssn, buf, len);
+			if (ssn->sslconn) {
+				r = socket_secure_write(ssn, buf, len);
 
-				if (w <= 0)
+				if (r <= 0)
 					goto fail;
-			} else
-#endif
-			{
-				w = write(ssn->socket, buf, len);
+			} else {
+				r = write(ssn->socket, buf, len);
 
-				if (w == -1) {
+				if (r == -1) {
 					error("writing data; %s\n",
 					    strerror(errno));
 					goto fail;
-				} else if (w == 0) {
+				} else if (r == 0) {
 					goto fail;
 				}
 			}
 
-			if (w > 0) {
-				len -= w;
-				buf += w;
-				wt += w;
+			if (r > 0) {
+				len -= r;
+				buf += r;
+				t += r;
 			}
 		}
 	}
@@ -384,7 +372,7 @@ socket_write(session *ssn, const char *buf, size_t len)
 		goto fail;
 	}
 
-	return wt;
+	return t;
 fail:
 	close_connection(ssn);
 
@@ -392,26 +380,24 @@ fail:
 }
 
 
-#ifndef NO_SSLTLS
 /*
  * Write data to a TLS/SSL connection.
  */
 ssize_t
 socket_secure_write(session *ssn, const char *buf, size_t len)
 {
-	int w, e;
+	int r, e;
 
 	for (;;) {
-		w = (ssize_t) SSL_write(ssn->ssl, buf, len);
-
-		if (w > 0)
+		if ((r = (ssize_t) SSL_write(ssn->sslconn, buf, len)) > 0)
 			break;
 
-		switch (SSL_get_error(ssn->ssl, w)) {
+		switch (SSL_get_error(ssn->sslconn, r)) {
 		case SSL_ERROR_ZERO_RETURN:
-			error("writing data; the connection has been closed "
-			    "cleanly\n");
-			return -1;
+			error("writing data through SSL; the connection has "
+			    "been closed cleanly\n");
+			goto fail;
+		case SSL_ERROR_NONE:
 		case SSL_ERROR_WANT_READ:
 		case SSL_ERROR_WANT_WRITE:
 		case SSL_ERROR_WANT_CONNECT:
@@ -420,21 +406,29 @@ socket_secure_write(session *ssn, const char *buf, size_t len)
 			break;
 		case SSL_ERROR_SYSCALL:
 			e = ERR_get_error();
-			if (e == 0)
-				error("writing data; EOF in violation of the "
-				    "protocol\n");
-			else if (e == -1)
-				error("writing data; %s\n", strerror(errno));
-			return -1;
+			if (e == 0 && r == 0)
+				error("writing data through SSL; EOF in "
+				    "violation of the protocol\n");
+			else if (e == 0 && r == -1)
+				error("writing data through SSL; %s\n",
+				    strerror(errno));
+			else
+				error("writing data through SSL; %s\n",
+				    ERR_error_string(e, NULL));
+			goto fail;
 		case SSL_ERROR_SSL:
-			error("writing data; %s\n",
+			error("writing data through SSL; %s\n",
 			    ERR_error_string(ERR_get_error(), NULL));
-			return -1;
+			goto fail;
 		default:
-			return -1;
+			break;
 		}
 	}
 
-	return w;
+	return r;
+fail:
+	SSL_set_shutdown(ssn->sslconn, SSL_SENT_SHUTDOWN |
+	    SSL_RECEIVED_SHUTDOWN);
+
+	return -1;
 }
-#endif
